@@ -3,7 +3,7 @@
 #'
 #'  Utilities for sequences, vectors, ranges of values
 #'
-#'       $Revision: 1.14 $ $Date: 2018/10/31 03:26:52 $
+#'       $Revision: 1.19 $ $Date: 2022/06/16 05:37:48 $
 #'
 #'  ==>>  ORIGINAL FILE is in spatstat/develop/Spatstat/R  <<==
 
@@ -26,9 +26,10 @@ orderwhich <- function(x, k, decreasing=FALSE) {
             index.return=TRUE)$ix[k]
 }
 
-## faster than sort(unique(x)) for numeric
-
-sortunique <- function(x) { rle(sort(x))$values }
+sortunique <- function(x) {
+  ## WAS FASTER: rle(sort(x))$values
+  sort(unique(x))
+}
 
 ## ................ reverse cumulative sum .....................
 
@@ -238,19 +239,18 @@ check.range <- function(r, fatal=TRUE) {
 }
 
 evenly.spaced <- function(x, tol=1e-07) {
-  # test whether x is evenly spaced and increasing
+  ## test whether x is evenly spaced and increasing
   dx <- diff(x)
   if(any(dx <= .Machine$double.eps))
     return(FALSE)
-  # The following test for equal spacing is used in hist.default
-  if(diff(range(dx)) > tol * mean(dx))
+  ## The following test for equal spacing is used in hist.default
+  if(length(dx) && (diff(range(dx)) > tol * mean(dx)))
     return(FALSE)
   return(TRUE)
 }
 
 equispaced <- function(z, reltol=0.001) {
-  dz <- diff(z)
-  return(diff(range(dz)) < reltol * mean(dz))
+  evenly.spaced(z, reltol)
 }
 
 
@@ -265,10 +265,9 @@ adjustthinrange <- function(ur,vstep,vr) {
 fastFindInterval <- function(x, b, labels=FALSE, reltol=0.001, dig.lab=3L) {
   nintervals <- length(b) - 1
   nx <- length(x)
-  if(nx == 0)
-    return(rep(0, nintervals))
-  ##
-  if(equispaced(b, reltol)) {
+  if(nx == 0) {
+    y <- integer(0)
+  } else if(equispaced(b, reltol)) {
     ## breaks are equally spaced
     zz <- .C(C_fastinterv,
              x          = as.double(x),
@@ -286,7 +285,7 @@ fastFindInterval <- function(x, b, labels=FALSE, reltol=0.001, dig.lab=3L) {
     #' digits in labels code copied from base::cut.default (with adaptations)
     for(dig in dig.lab:max(12L, dig.lab)) {
       ch.br <- formatC(0 + b, digits = dig, width = 1L)
-      if(ok <- all(ch.br[-1L] != ch.br[1L:nintervals]))
+      if(all(ch.br[-1L] != ch.br[1L:nintervals]))
         break
     }
     blab <- paste0("[",
@@ -406,4 +405,58 @@ niceround <- function(x, m=c(1,2,5,10)) {
   return(z)
 }
 
+## ..............................................
 
+exactCutBreaks <- function(x, breaks) {
+  ## determine the exact breakpoints used in cut.default
+  ## This code is extracted from base::cut.default
+  stopifnot(is.numeric(x))
+  if(length(breaks) > 1L) {
+    ## numeric vector of breaks
+    breaks <- sort.int(as.double(breaks))
+    if(anyDuplicated(breaks)) 
+      stop("'breaks' are not unique")
+  } else if(length(breaks) == 1L) {
+    ## number of breaks
+    if (is.na(breaks) || breaks < 2L) 
+      stop("invalid number of intervals")
+    nb <- as.integer(breaks + 1)
+    dx <- diff(rx <- range(x, na.rm = TRUE))
+    if(dx == 0) {
+      dx <- if(rx[1L] != 0) abs(rx[1L]) else 1
+      breaks <- seq.int(rx[1L] - dx/1000,
+                        rx[2L] + dx/1000, 
+                        length.out = nb)
+    } else {
+      breaks <- seq.int(rx[1L], rx[2L], length.out = nb)
+      breaks[c(1L, nb)] <- c(rx[1L] - dx/1000,
+                             rx[2L] + dx/1000)
+    }
+  } else stop("breaks must be specified")
+  return(breaks)
+}
+
+
+## ....  Harmonic mean ..........
+
+harmonicmean <- function(x, na.rm=TRUE) {
+  ## harmonic mean, robust against zeroes and small values
+  if(anyNA(x)) {
+    if(na.rm) x <- x[!is.na(x)] else return(NA)
+  }
+  if(length(x) == 0) return(NaN) # consistent with mean()
+  mx <- min(abs(x))
+  if(mx == 0) return(mx)
+  return(mx/mean(mx/x))
+}
+
+harmonicsum <- function(x, na.rm=TRUE) {
+  # computes 1/sum(1/x) robustly
+  if(anyNA(x)) {
+    if(na.rm) x <- x[!is.na(x)] else return(NA)
+  }
+  if(length(x) == 0) return(0) # consistent with sum()
+  mx <- min(abs(x))
+  if(mx == 0) return(mx)
+  return(mx/sum(mx/x))
+}
